"""
web_server.py

spins up a flask web server to host an api
"""


from flask import Flask, send_file, request
import zipfile 
import io
from .gen_keybinding_img import make_imgs 
from .gen_keybinding_img import this_dir as img_dir
from .discord_log import WALLPAPER_PATH
# from threading import Thread
import shutil
from multiprocessing import Process 
from os.path import expanduser, islink
import os
from libqtile import hook
from libqtile.log_utils import logger
from .auto_desk_api import set_layout, find_layout_file
from os import remove as rm, symlink
from .tmux import tmux_layout
from libqtile.command.client import InteractiveCommandClient
import alsaaudio
from libqtile.command.base import SelectError, CommandError
from gi import require_version

# from gi.repository import Playerctl, Gio
from gi.repository.GLib import GError


app = Flask("frankentile")
API_HANDLE = None


@app.route("/key-binds", methods=["GET"])
def key_binds():
    """sends back a key bindings image"""
    try:
        shutil.rmtree(img_dir)
    except FileNotFoundError:
        pass

    images = make_imgs(config_path=expanduser("~/.config/qtile/config.py"))

    zip_f = os.path.join(img_dir, "keybinds.zip")

    with zipfile.ZipFile(zip_f, "w", zipfile.ZIP_DEFLATED, False) as zip_file:
        for fname in images:
            zip_file.writestr(fname.split("/")[-1], open(fname, "rb").read())

    return send_file(zip_f, download_name="keybinds.zip", as_attachment=True)


@app.route("/wallpaper", methods=["POST"])
def set_wallpaper():
    """takes a path to an image and sets the wallpaper to it"""
    path = request.values.get("wallpaper-path")

    if islink(WALLPAPER_PATH):
        rm(WALLPAPER_PATH)
        symlink(expanduser(path), WALLPAPER_PATH)

    # TODO: reload configs

    return "success"


@app.route("/auto-desk/layout/<layout>")
def load_layout(layout):
    """uses auto-desk to load the specified layout"""
    # TODO: check for layout in "~/.config/auto-desk/layouts/" (ignore spaces, hyphens, and underscores)
    layout_name = find_layout_file(layout)

    if layout_name: 
        res = set_layout(layout_name)

        return f"setting auto-desk layout {layout}. auto-desk says: {res}"
    else:
        return f"unknown layout: {layout}" 


@app.route("/tmux/<layout>")
async def tmux(layout):
    """sets up a tmux layout"""
    res = await tmux_layout(layout)

    return f"setting up auto-tmux layout {layout}. auto-tmux says: {res}"


@app.route("/focus-on/<group>")
def focus_on(group: str):
    """changes active focus to the specified group"""
    c = InteractiveCommandClient()
    
    try:
        c.group[group].toscreen()
    except SelectError:
        return "no group by that name"
    else:
        return "focus shifted"


@app.route("/move-to/<group>")
def move_to(group: str):
    """move current window to group"""
    c = InteractiveCommandClient()
    
    try:
        c.window.togroup(group)
    except CommandError:
        return "no group by that name"
    else:
        return "window moved" 


@app.route("/track/<control>")
def music(control: str):
    """used to control the music (play/pause/next/etc)"""
    # TODO: make the play command open a player then play (if no player is open)
    try:
        require_version('Playerctl', '2.0')
    except ValueError as e:
        return f"setting name space return error: {e}"

    from gi.repository import Playerctl

    player = Playerctl.Player()

    controls = {
        "play": player.play, 
        "pause": player.pause,
        "play-pause": player.play_pause,
        "next": player.next,
        "prev": player.previous,
    }

    f = controls.get(control)

    if not f:
        return f"no track control command by the name \"{control}\" found. must be one of, {list(controls.keys())}"

    try:
        f()
    except GError:
        return "track control failed, is there a music player running?"
    else:
        return "track controlled succefully"


@app.route("/volume/<cmd>")
def volume(cmd):
    """adjusts volume"""
    m = alsaaudio.Mixer()
    vol = m.getvolume()
    vol = int(vol[0])
        
    volumes = {
        "up": vol + 5,
        "down": vol - 5,
        "mute": 0,
        }

    new_vol = volumes.get(cmd)

    if new_vol is None:
        return f"\"{cmd}\" is not a valid volume control command. please use one of {list(volumes.keys())}"

    m.setvolume(new_vol)
    return "volume adjusted correctly"


@app.route("/sleep")
def sleep():
    """puts the computer to sleep using systemctl"""
    import os

    os.system("systemctl suspend")


@app.route("/config-reload")
def config_reload():
    """reloads qtile configs"""
    c = InteractiveCommandClient()
    c.reload_config()

    return "configs reloaded succefully"


def start_app(host, port):
    logger.warning("running web server")
    app.run(host=host, port=port)  # , debug=True, passthrough_errors=True, use_debugger=False, use_reloader=True)


def start_api(host="127.0.0.1", port=8080):
    """starts the flask server"""
    p = Process(target=start_app, args=[host, port])
    p.start()
    global API_HANDLE
    API_HANDLE = p


def stop_api():
    if API_HANDLE:
        API_HANDLE.terminate()


def init():
    hook.subscribe.startup_once(start_api)
    hook.subscribe.shutdown(stop_api)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
                    prog='frankentile web',
                    description='starts a web server to control Qtile.',
                    )
    parser.add_argument('ip_adr')

    args = parser.parse_args()

    start_app(args.ip_adr, 8080)
